@import "dark_colours.css";
@import "dark_buttons.css";


label.info {
    background-color: @signal-info;
    background-image: image(@signal-info);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

label.error {
    background-color: @signal-danger;
    background-image: image(@signal-danger);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

label.success {
    background-color: @signal-success;
    background-image: image(@signal-success);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

#vpn-widget,
#login-widget {
    margin: 10px;
}

#login-logo {
    margin-top: 75px;
    margin-bottom: 40px;
    margin-right: 10px;
}

#two-factor-auth-vpn-logo {
    margin-top: 2em;
    margin-bottom: 1em;
}

#bug-report-content,
#bug-report-dialog buttonbox,
#early-access-dialog buttonbox,
.dialog-action-area{
    margin: 20px;
}
.dialog-action-area > button {
    margin-left: 5px;
}

#server-load.signal-danger {
    color: @signal-danger;
}

#server-load.signal-warning {
    color: @signal-warning;
}

#server-load.signal-success {
    color: @signal-success;
}

.setting-category {
    margin-top: 20px;
    margin-left: 20px;
    margin-right: 20px;
}

.setting-category:last-child {
    margin-bottom: 60px;
}

.setting-item {
    border: 1px solid rgba(155, 155, 155, 0.2);
    padding: 10px;
    border-radius: 10px;
}

.upgrade-tag {
    border-radius: 10px;
    background: transparent;
    border: 1px solid rgba(155, 155, 155, 0.2);
    outline: none;
    background-image: none;
}

.release-notes-collection {
    margin: 20px 20px 60px 20px;
}

#login-kill-switch-revealer box {
    border-top: 1px solid rgba(155, 155, 155, 0.2);
    padding: 15px 0px;
    margin-top: 20px;
}

label.signal-danger {
    color: @signal-danger;
}

#connection-status-label {
    font-weight: bold;
}

#port-forwarding-widget {
    background-image: none;
    transition: background-color 0.2s;
    border-radius: 5px;
}

#port-forwarding-widget:hover {
    background-image: none;
    background-color: rgba(255, 255, 255, 0.08);
}

#port-forwarding-widget:active {
    background-image: none;
    background-color: rgba(255, 255, 255, 0.02);
}

#port-forwarding-widget>box {
    padding: 5px;
}

#split-tunneling-app-selection-window viewport {
    margin: 15px 30px 30px;
    padding: 2px 0px 0px 0px;
    border: 1px solid @border-norm;
    border-radius: 5px;
}

#split-tunneling-app-selection-window viewport > box {
    margin: 0px 0px 0px 5px;
}

#split-tunneling-app-selection-window #split-tunneling-app-done-button {
    margin: 15px 30px 30px;
}

#split-tunneling-app-based-settings {
    margin-top: 10px;
}

#split-tunneling-app-based-settings #split-tunneling-app-add-button {
    margin: 15px 10px 5px 10px;
}


#selected-app-list {
    padding: 12px 16px 12px 16px;
    border-radius: 5px;
    border: 1px solid rgba(155, 155, 155, 0.2);
}

#two-factor-auth-widget > label.two-factor-auth-stack-title {
    font-size: 1.3em;
    font-weight: bold;
    margin-bottom: 0.5em;
}

.default-loading-widget-label{
    margin-left: 1.8em;
	margin-right: 1.8em;
}


/* Show under maintenance icon vertically aligned with toggle button icons */
.row-content .under-maintenance-icon {
    margin: 10px;
}

/* Keep row height consistent whether showing connect button or upgrade link */
.row-content {
    min-height: 32px;
}

#feature-legend-popover {
    padding: 0;
}

#feature-legend-popover .feature-legend-header {
    margin: 6px 10px;
    border-spacing: 6px;
}

#feature-legend-popover .feature-legend-content {
    margin: 0 10px 10px 10px;
    padding-top: 2px;
    border-spacing: 10px;
}

#feature-legend-popover .feature-legend-row {
    border-spacing: 6px;
}

#feature-legend-popover .feature-legend-icon {
    margin-top: 3px;
}

#feature-legend-popover .feature-legend-learn-more {
    padding: 0;
}

#server-list-header-row {
    border-spacing: 10px;
}

#nps-survey-modal-content {
    margin: 24px 32px 32px 32px;
}

#nps-survey-modal.submitted #nps-survey-modal-content {
    margin-top: 0;
}

#nps-survey-modal headerbar {
    background: transparent;
    background-image: none;
    box-shadow: none;
    border: none;
    min-height: 32px;
    padding: 0 4px;
}

.nps-score-layout {
    margin-top: 32px;
}

.nps-score-button {
    padding: 0;
    min-width: 32px;
    min-height: 32px;
    border: none;
    outline: none;
    border-radius: 4px;
}

.nps-score-button + .nps-score-button {
    margin-left: 4px;
}

.nps-score-button:checked {
    background: @interaction-norm-active;
    background-image: none;
    color: @text-norm;
}

.nps-light-label {
    font-size: 12px;
    font-weight: 400;
    margin: 0;
}

.nps-feedback-optional-label {
    margin-top: 4px;
}

.nps-feedback-prompt-label {
    font-size: 12px;
    font-weight: 700;
    margin-top: 32px;
}

.nps-feedback-scroll {
    margin-top: 8px;
}

#nps-survey-modal-title-label {
    font-size: 20px;
    font-weight: 700;
    margin: 32px 0 0 0;
}

#nps-survey-modal button.primary {
    margin-top: 24px;
    margin-bottom: 32px;
}

#nps-survey-modal-subtitle-label {
    font-size: 14px;
    font-weight: 400;
    margin: 12px 0 0 0;
}